# File generated with docstub

import numpy as np
from xarray import DataTree

from arviz_base import from_dict

def generate_base_data(seed=...) -> None: ...
def datatree(seed=...) -> None: ...
def datatree2(seed=...) -> None: ...
def datatree3(seed=...) -> None: ...
def datatree_binary(seed=...) -> None: ...
def datatree_regression(seed=...) -> None: ...
def datatree_4d(seed=...) -> None: ...
def datatree_sample(seed=...) -> None: ...
def cmp() -> None: ...
def fake_dt() -> None: ...
def datatree_censored(seed=...) -> None: ...
def check_multiple_attrs(
    test_dict: dict[str, list[str]], parent: DataTree
) -> list[str | tuple[str, str]]: ...
