# File generated with docstub

from collections.abc import Hashable, Mapping, Sequence
from typing import TYPE_CHECKING, Any

from _typeshed import Incomplete

if TYPE_CHECKING:
    from numpy.typing import ArrayLike, NDArray

CoordSpec: Incomplete
CoordOut: Incomplete
DimSpec: Incomplete
DimOut: Incomplete

DictData: Incomplete
